//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: Follower.cs $ $Revision: 6 $
//-----------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Xml;

using System.Windows.Forms;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;
using Microsoft.Ccr.Adapters.WinForms;

using W3C.Soap;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using Microsoft.Robotics.PhysicalModel;

using bumper = Microsoft.Robotics.Services.ContactSensor.Proxy;
using drive = Microsoft.Robotics.Services.Drive.Proxy;
using sicklrf = Microsoft.Robotics.Services.Sensors.SickLRF.Proxy;

using winMsg = Microsoft.Robotics.Services.Sample.WinMessage.Proxy;

using vision = Microsoft.Robotics.Services.Sample.SimpleVision.Proxy;
using follower = Microsoft.Robotics.Services.Sample.Follower;

using Microsoft.Robotics.Services.Sample.Follower.Properties;

namespace Microsoft.Robotics.Services.Sample.Follower
{

    /// <summary>
    /// Implementation class for Follower
    /// </summary>
    [DisplayName("Follower")]
    [Description("Provides access to a set of services that support simple visual tracking, speech recognition, text-to-speech, and obstacle avoidance.")]
    [Contract(Contract.Identifier)]
    public class FollowerService : DsspServiceBase
    {
        /// <summary>
        /// _state
        /// </summary>
        private FollowerState _state = new FollowerState();

        /// <summary>
        /// _main Port
        /// </summary>
        [ServicePort("/follower", AllowMultipleInstances = false)]
        private FollowerOperations _mainPort = new FollowerOperations();

        private Port<DisableSpeechUpdate> _disableSpeechPort = new Port<DisableSpeechUpdate>();
        // port for periodic timer messages used to call TimerHandler
        private Port<DateTime> _timerPort = new Port<DateTime>();

        #region partners
        [Partner("Bumper", Contract = bumper.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        bumper.ContactSensorArrayOperations _bumperPort = new bumper.ContactSensorArrayOperations();
        bumper.ContactSensorArrayOperations _bumperNotify = new bumper.ContactSensorArrayOperations();

        [Partner("Drive", Contract = drive.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        drive.DriveOperations _drivePort = new drive.DriveOperations();
        drive.DriveOperations _driveNotify = new drive.DriveOperations();

        //[Partner("Laser", Contract = sicklrf.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExisting)]
        //sicklrf.SickLRFOperations _laserPort = new sicklrf.SickLRFOperations();
        //sicklrf.SickLRFOperations _laserNotify = new sicklrf.SickLRFOperations();

        [Partner("WINMSG", Contract = winMsg.Contract.Identifier, CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        winMsg.WinMessageOperations _winMsgPort = new winMsg.WinMessageOperations();
        winMsg.WinMessageOperations _winMsgNotify = new winMsg.WinMessageOperations();

        [Partner("Vision", Contract = vision.Contract.Identifier, Optional = false,
           CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        vision.SimpleVisionOperations _visionPort = new vision.SimpleVisionOperations();
        vision.SimpleVisionOperations _visionNotify = new vision.SimpleVisionOperations();
        #endregion

        #region constants
        // The follower uses there constants to evaluate open space 
        // and adjust speed and rotation of the differential drive.
        // Change those values to match your robot.
        // You could expose them on the follower state type and make the
        // service configurable.

        /// <summary>
        /// Amount to backup when hitting an obstacle.
        /// </summary>
        const int BackupDistance = -150; // mm

        /// <summary>
        /// If an obstacle comes within thisdistance the robot stops moving.
        /// </summary>
        const int ObstacleDistance = 450; // mm

        /// <summary>
        /// If the robot is mapping and has this much open space ahead it stops mapping
        /// and enters the space.
        /// </summary>
        const int SafeDistance = 2000; // mm

        /// <summary>
        /// The width of the corridor that must be safe in order to got from mapping to moving.
        /// </summary>
        const int CorridorWidthMapping = 350; // mm

        /// <summary>
        /// The minimum free distance that is required to drive with max. velocity.
        /// </summary>
        const int FreeDistance = 3000; // mm

        /// <summary>
        /// If the free space is at least this distance the robot operates at 1/2 max. velocity otherwise 
        /// the robot slows down to 1/4 of max. vel.
        /// </summary>
        const int AwareOfObstacleDistance = 1500; // mm

        /// <summary>
        /// The max. velocity with which to move.
        /// </summary>
        const int MaximumForwardVelocity = 1000; // mm/sec

        /// <summary>
        /// The with of the corridor in which obstacles effect velocity.
        /// </summary>
        const int CorridorWidthMoving = 500; // mm
        #endregion

        /// <summary>
        /// Default Service Constructor
        /// </summary>
        public FollowerService(DsspServiceCreationPort creationPort)
            :
                base(creationPort)
        {
        }

        /// <summary>
        /// Service Start
        /// </summary>
        protected override void Start()
        {
            base.Start();

            MainPortInterleave.CombineWith(
                Arbiter.Interleave(
                    new TeardownReceiverGroup(),
                    new ExclusiveReceiverGroup(
                        Arbiter.Receive<DateTime>(true, _timerPort, TimerHandler),
                        Arbiter.Receive<DisableSpeechUpdate>(true, _disableSpeechPort, DisableSpeechHandler),
                        Arbiter.Receive<winMsg.ReceiveMessage>(true, _winMsgNotify, OnSpeechRecognition),
                        Arbiter.Receive<vision.NotifyObjectDetection>(true, _visionNotify, OnObjectDetection),
                        Arbiter.Receive<vision.NotifyFaceDetection>(true, _visionNotify, OnFaceDetection),
                        Arbiter.Receive<vision.NotifyHandGestureDetection>(true, _visionNotify, OnHandGestureDetection)
                    ),
                    new ConcurrentReceiverGroup(
                        Arbiter.Receive<bumper.Replace>(true, _bumperNotify, BumperReplaceNotification),
                        Arbiter.Receive<bumper.Update>(true, _bumperNotify, BumperUpdateNotification),
                        //Arbiter.Receive<sicklrf.Reset>(true, _laserNotify, LaserResetNotification),
                        Arbiter.Receive<drive.Update>(true, _driveNotify, DriveUpdateNotification)
                    )
                )
            );

            //Activate(Arbiter.Receive<sicklrf.Replace>(false, _laserNotify, LaserReplaceNotification));

            // Create Subscriptions
            _bumperPort.Subscribe(_bumperNotify);
            _drivePort.Subscribe(_driveNotify);
            //_laserPort.Subscribe(_laserNotify);
            _winMsgPort.Subscribe(_winMsgNotify);
            _visionPort.Subscribe(_visionNotify);

            WinFormsServicePort.Post(new RunForm(CreateForm));
            WinFormsServicePort.FormInvoke(
                delegate()
                {
                    _form.UpdateJoystickAxes(new Axes());
                }
            );

            _state.TimerIntervalInMs = 150;
            TaskQueue.EnqueueTimer(TimeSpan.FromMilliseconds(_state.TimerIntervalInMs), _timerPort);
        }

        MyForm _form;
        Form CreateForm()
        {
            _form = new MyForm(_mainPort, _drivePort);
            return _form;
        }

        /// <summary>
        /// Get Handler
        /// </summary>
        /// <param name="get"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public virtual IEnumerator<ITask> GetHandler(Get get)
        {
            get.ResponsePort.Post(_state);
            yield break;
        }

        /// <summary>
        /// Drop Handler
        /// </summary>
        /// <param name="drop"></param>
        /// <returns></returns>
        [ServiceHandler(ServiceHandlerBehavior.Teardown)]
        public virtual IEnumerator<ITask> DropHandler(DsspDefaultDrop drop)
        {
            StopMoving();
            
            WinFormsServicePort.Post(new Shutdown());

            if (_visionPort != null)
            {
                _visionPort.DsspDefaultDrop();
                _visionPort = null;
            }

            base.DefaultDropHandler(drop);
            _state = null;
            yield break;
        }

        /// <summary>
        /// SetMotorActivate Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SetMotorActivateHandler(SetMotorActivate replace)
        {
            if (_state.DriveState == null || !_state.DriveState.IsEnabled)
            {
                EnableMotor(true);
            }
            
            disableMotor = replace.Body.OnOff;

            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
            yield break;
        }

        /// <summary>
        /// SetBehaviorState Handler
        /// </summary>
        [ServiceHandler(ServiceHandlerBehavior.Concurrent)]
        public IEnumerator<ITask> SetBehaviorStateHandler(SetRobotBehavior update)
        {
            SetRobotBehavior(update.Body.Behavior);

            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
            yield break;
        }

        bool disableSpeech = false;

        void DisableSpeechHandler(DisableSpeechUpdate update)
        {
            disableSpeech = update.Body.OnOff;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// TimerHandler
        /// </summary>
        /// <returns></returns>
        void TimerHandler(DateTime timestamp)
        {
            //Console.WriteLine("timer");
            if (_state == null)
            {
                return;
            }
            try
            {
                if (_state.currentObjectResult == null)
                    _state.currentObjectResult = new vision.ObjectResult();

                if (_state.currentFaceResult == null)
                    _state.currentFaceResult = new vision.FaceResult();

                if (_state.currentHandGestureResult == null)
                    _state.currentHandGestureResult = new vision.HandGestureResult();

                // update robot state here
                UpdateLogicalState();
                UpdateRobotState();

                if (_state.FaceResultValidTime > 0)
                    _state.FaceResultValidTime--;

                if (_state.HandGestureResultValidTime > 0)
                    _state.HandGestureResultValidTime--;

                if (_state.SpeechCommandValidTime > 0)
                    _state.SpeechCommandValidTime--;

            }
            catch (Exception ex)
            {
                LogError(LogGroups.Console, "Error in UpdateRobotState ", ex);
            }

            TaskQueue.EnqueueTimer(TimeSpan.FromMilliseconds(_state.TimerIntervalInMs), _timerPort);
        }

        const int ImageWidth = vision.SimpleVisionState.ImageWidth;
        const int ImageHeight = vision.SimpleVisionState.ImageHeight;

        void OnObjectDetection(vision.NotifyObjectDetection update)
        {
            _state.currentObjectResult = update.Body;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        void OnFaceDetection(vision.NotifyFaceDetection update)
        {
            _state.currentFaceResult = update.Body;

            //LogInfo("OnFaceDetection");
            _state.FaceResultValidTime = 10;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        void OnHandGestureDetection(vision.NotifyHandGestureDetection update)
        {
            _state.currentHandGestureResult = update.Body;

            //LogInfo("OnHandGestureDetection");
            _state.HandGestureResultValidTime = 10;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// Handles Update notification from the Drive partner
        /// </summary>
        /// <param name="update">notification</param>
        void DriveUpdateNotification(drive.Update update)
        {
            _state.DriveState = update.Body;
            _state.Velocity = (VelocityFromWheel(update.Body.LeftWheel) + VelocityFromWheel(update.Body.RightWheel)) / 2;
            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        /// <summary>
        /// Computes the wheel velocity in mm/s.
        /// </summary>
        /// <param name="wheel">wheel</param>
        /// <returns>velocity</returns>
        private static int VelocityFromWheel(Microsoft.Robotics.Services.Motor.Proxy.WheeledMotorState wheel)
        {
            if (wheel == null)
            {
                return 0;
            }
            return (int)(1000 * wheel.WheelSpeed); // meters to millimeters
        }

        #region  Bumper handlers

        bool simulatedBumper = false;

        /// <summary>
        /// Handles Replace notifications from the Bumper partner
        /// </summary>
        /// <param name="replace">notification</param>
        void BumperReplaceNotification(bumper.Replace replace)
        {
            bumper.ContactSensorArrayState bumpers = replace.Body;
            if (bumpers.Sensors == null)
                return;
            foreach (bumper.ContactSensor s in bumpers.Sensors)
            {
                if (s.Pressed)
                {
                    simulatedBumper = true;
                    if (s.HardwareIdentifier == 0)
                        _state.BumperPressed = BumperPressedType.Front;
                    else
                        _state.BumperPressed = BumperPressedType.Rear;

                    LogInfo("BumperPressed = " + s.HardwareIdentifier);
                }
            }

            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);
        }

        /// <summary>
        /// Handles Update notification from the Bumper partner
        /// </summary>
        /// <param name="update">notification</param>
        void BumperUpdateNotification(bumper.Update update)
        {
            if (simulatedBumper) return;
            if (update.Body.Pressed)
            {
                //if (update.Body.HardwareIdentifier < 5)
                    _state.BumperPressed = BumperPressedType.Front;
                //else
                //    _state.BumperPressed = BumperPressedType.Rear;

                LogInfo("BumperPressed = " + update.Body.HardwareIdentifier);
            }

            update.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }

        #endregion

        #region Laser handlers

        /// <summary>
        /// Handles Replace notifications from the Laser partner
        /// </summary>
        /// <param name="replace">notification</param>
        void LaserReplaceNotification(sicklrf.Replace replace)
        {
            // When this handler is called a couple of notifications may
            // have piled up. We only want the most recent one.
            sicklrf.State laserData = GetMostRecentLaserNotification(replace.Body);

            _state.MostRecentLaserData = laserData;
            _state.MostRecentLaserTime = laserData.TimeStamp;

            replace.ResponsePort.Post(DefaultReplaceResponseType.Instance);

            // Skip messages that have been queued up in the meantime.
            // The notification that are lingering are out of data by now.
            GetMostRecentLaserNotification(laserData);

            // Reactivate the handler.
            //Activate(Arbiter.Receive<sicklrf.Replace>(false, _laserNotify, LaserReplaceNotification));
        }

        /// <summary>
        /// Gets the most recent laser notification. Older notifications are dropped.
        /// </summary>
        /// <param name="laserData">last known laser data</param>
        /// <returns>most recent laser data</returns>
        private sicklrf.State GetMostRecentLaserNotification(sicklrf.State laserData)
        {
            //sicklrf.Replace testReplace;

            // _laserNotify is a PortSet<>, P3 represents IPort<sicklrf.Replace> that
            // the portset contains
            int count = 0;//_laserNotify.P3.ItemCount - 1;

            for (int i = 0; i < count; i++)
            {
                //testReplace = _laserNotify.Test<sicklrf.Replace>();
                //if (testReplace.Body.TimeStamp > laserData.TimeStamp)
                //{
                //    laserData = testReplace.Body;
                //}
            }

            if (count > 0)
            {
                LogInfo("Dropped " + count + " laser readings (laser start)");
            }
            return laserData;
        }

        /// <summary>
        /// Handles the reset notification of the Laser partner.
        /// </summary>
        /// <param name="reset">notification</param>
        void LaserResetNotification(sicklrf.Reset reset)
        {
            if (_state.LogicalState != LogicalStateType.Stop)
            {
                LogInfo("Stop requested: laser reported reset");
                StopMoving();

            }
            reset.ResponsePort.Post(DefaultSubmitResponseType.Instance);
        }

        /// <summary>
        /// Respresent a laser range finder reading
        /// </summary>
        class RangeData
        {
            /// <summary>
            /// Creates a new instance.
            /// </summary>
            /// <param name="distance">measured distance</param>
            /// <param name="heading">heading in degrees</param>
            public RangeData(int distance, double heading)
            {
                _distance = distance;
                _heading = heading;
            }

            int _distance;
            double _heading;

            /// <summary>
            /// Gets the distance in milimeters.
            /// </summary>
            public int Distance
            {
                get { return _distance; }
            }

            /// <summary>
            /// Gets the heading in degrees.
            /// </summary>
            public double Heading
            {
                get { return _heading; }
            }

            /// <summary>
            /// Comparer to sort instances by distance.
            /// </summary>
            /// <param name="first">first reading</param>
            /// <param name="second">second reading</param>
            /// <returns>a value les than 0 if  <paramref name="first"/> is closer than <paramref name="second"/>, 0 if both have the same distance, a value greater 0 otherwise</returns>
            public static int ByDistance(RangeData first, RangeData second)
            {
                return first._distance.CompareTo(second._distance);
            }
        }


        /// <summary>
        /// Finds the best heading in a 180 degree laser scan
        /// </summary>
        /// <param name="laserData">laser scan</param>
        /// <param name="dx">horizontal offset</param>
        /// <param name="dy">vertical offset</param>
        /// <param name="width">width of corridor that must be free</param>
        /// <returns>best heading in degrees</returns>
        private static int FindBestFrom(sicklrf.State laserData, int dx, int dy, int width)
        {
            int count = laserData.DistanceMeasurements.Length;
            double span = Math.PI * laserData.AngularRange / 180.0;

            List<RangeData> ranges = new List<RangeData>();

            for (int i = 0; i < count; i++)
            {
                int range = laserData.DistanceMeasurements[i];
                double angle = span * i / count - span / 2.0;

                double x = range * Math.Sin(angle) - dx;
                double y = range * Math.Cos(angle) - dy;

                angle = Math.Atan2(-x, y);
                range = (int)Math.Sqrt(x * x + y * y);

                ranges.Add(new RangeData(range, angle));
            }

            ranges.Sort(RangeData.ByDistance);

            for (int i = 0; i < ranges.Count; i++)
            {
                RangeData curr = ranges[i];

                double delta = Math.Atan2(width, curr.Distance);
                double low = curr.Heading - delta;
                double high = curr.Heading + delta;

                for (int j = i + 1; j < ranges.Count; j++)
                {
                    if (ranges[j].Heading > low &&
                        ranges[j].Heading < high)
                    {
                        ranges.RemoveAt(j);
                        j--;
                    }

                }
            }

            ranges.Reverse();

            int bestDistance = ranges[0].Distance;
            double bestHeading = ranges[0].Heading;
            Random rand = new Random();

            for (int i = 0; i < ranges.Count; i++)
            {
                if (ranges[i].Distance < bestDistance)
                {
                    break;
                }
                if (rand.Next(i + 1) == 0)
                {
                    bestHeading = ranges[i].Heading;
                }
            }

            return -(int)Math.Round(180 * bestHeading / Math.PI);
        }

        /// <summary>
        /// Finds closest obstacle in a corridor.
        /// </summary>
        /// <param name="laserData">laser scan</param>
        /// <param name="width">corridor width</param>
        /// <param name="fov">field of view in degrees</param>
        /// <returns>distance to the closest obstacle</returns>
        private static int FindNearestObstacleInCorridor(sicklrf.State laserData, int width, int fov)
        {
            // Fixed null exception of sicklrf.State in simulation
            if (laserData == null) return 0;
            if (laserData.DistanceMeasurements == null) return 0;

            int index;
            int best = 8192;
            int count = laserData.DistanceMeasurements.Length;
            double rangeLow = -laserData.AngularRange / 2.0;
            //double rangeHigh = laserData.AngularRange / 2.0;
            double span = laserData.AngularRange;

            for (index = 0; index < count; index++)
            {
                double angle = rangeLow + (span * index) / count;
                if (Math.Abs(angle) < fov)
                {
                    angle = angle * Math.PI / 180;

                    int range = laserData.DistanceMeasurements[index];

                    // skip if range is 0 in simulatedLRF
                    if (range == 0) continue;

                    int x = (int)(range * Math.Sin(angle));
                    //int y = (int)(range * Math.Cos(angle));

                    if (Math.Abs(x) < width)
                    {
                        if (range < best)
                        {
                            best = range;
                        }
                    }
                }
            }

            return best;
        }
        #endregion

        #region Drive helper method
        /// <summary>
        /// disableMotor flag for a robot activation.
        /// </summary>
        public bool disableMotor = true;

        /// <summary>
        /// Sets the forward velocity of the drive.
        /// </summary>
        /// <param name="speed">velocity in mm/s</param>
        void Move(int speed)
        {
            if (_state.DriveState == null || !_state.DriveState.IsEnabled)
            {
                EnableMotor(true);
            }

            if (speed == 0)
                _state.LogicalState = LogicalStateType.Stop;
            else
                _state.LogicalState = LogicalStateType.Move;

            drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
            request.LeftWheelPower = (double)speed / 1000.0;
            request.RightWheelPower = (double)speed / 1000.0;

            if (disableMotor != true)
            {
                //LogInfo(string.Format("Move: speed(mm/s) = {0}", speed));
                _drivePort.SetDrivePower(request);
            }
            return;
        }

        /// <summary>
        /// Moves the drive relative to its Adjusted heading.
        /// </summary>
        /// <param name="angle">angle in degrees</param>
        /// <param name="step">distance</param>
        void AdjustedMove(int angle, int step)
        {
            if (_state.DriveState == null || !_state.DriveState.IsEnabled)
            {
                EnableMotor(true);
            }

            double Degrees = (double)(-angle);

            double rightPow = 0;
            double leftPow = 0;
            double power = 100;

            _state.LogicalState = LogicalStateType.AdjustedMove;

            if (Math.Abs(Degrees) > 30)
                power = power / 1.5;

            if (Degrees < 0)
            {
                rightPow = power;
                leftPow = power * (1 - Math.Abs((Degrees % 90) / 90));
            }
            else
            {
                leftPow = power;
                rightPow = power * (1 - Math.Abs((Degrees % 90)) / 90);
            }

            _state.Countdown = (int)Math.Abs(step / 10);

            drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
            request.LeftWheelPower = (double)leftPow / 1000.0;
            request.RightWheelPower = (double)rightPow / 1000.0;

            if (disableMotor != true)
            {
                //LogInfo(string.Format("AdjustedMove: (mm/s) rightPow {0} leftPow {1}", rightPow, leftPow));
                _drivePort.SetDrivePower(request);
            }
            return;
        }

        /// <summary>
        /// Moves the drive forward for the specified distance.
        /// </summary>
        /// <param name="step">distance</param>
        void Translate(int step)
        {
            if (_state.DriveState == null || !_state.DriveState.IsEnabled)
            {
                EnableMotor(true);
            }

            _state.LogicalState = LogicalStateType.Translate;
            _state.Countdown = (int)Math.Abs(step / 30);
            double sign = 1;
            if (step < 0)
                sign = -1;
            double power = sign * 120;
            drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
            request.LeftWheelPower = (double)power / 1000.0;
            request.RightWheelPower = (double)power / 1000.0;

            if (disableMotor != true)
            {
                //LogInfo(string.Format("Translate: step(mm) {0}", step));
                _drivePort.SetDrivePower(request);
            }
            return;
        }


        /// <summary>
        /// Turns the drive relative to its current heading.
        /// </summary>
        /// <param name="angle">angle in degrees</param>
        /// <param name="turnType">TurnType</param>
        void Turn(int angle, TurnType turnType)
        {
            if (_state.DriveState == null || !_state.DriveState.IsEnabled)
            {
                EnableMotor(true);
            }

            double Degrees = (double)(-angle);

            double rightPow = 0;
            double leftPow = 0;
            double power = 70;

            _state.LogicalState = LogicalStateType.Turn;

            if (Math.Abs(Degrees) < 20)
                power = power / 2;

            if (turnType == TurnType.OneWheelTurn)
            {
                if (Degrees < 0)
                {
                    rightPow = power * 1.5;
                }
                else
                {
                    leftPow = power * 1.5;
                }
                _state.Countdown = (int)Math.Abs(angle / 15);
            }
            else
            {
                if (Degrees < 0)
                {
                    rightPow = power;
                    leftPow = -power;
                }
                else
                {
                    rightPow = -power;
                    leftPow = power;
                }
                _state.Countdown = (int)Math.Abs(angle / 20);
            }

            drive.SetDrivePowerRequest request = new drive.SetDrivePowerRequest();
            request.LeftWheelPower = (double)leftPow / 1000.0;
            request.RightWheelPower = (double)rightPow / 1000.0;

            if (disableMotor != true)
            {
                //LogInfo(string.Format("Turn: (mm/s) rightPow {0} leftPow {1}", rightPow, leftPow));
                _drivePort.SetDrivePower(request);
            }
            return;
        }

        /// <summary>
        /// Sets the velocity of the drive to 0.
        /// </summary>
        void StopMoving()
        {
            Move(0);
        }

        /// <summary>
        /// Sets the drives enabled state.
        /// </summary>
        /// <param name="enable">new enables state</param>
        /// <returns>response port</returns>
        PortSet<DefaultUpdateResponseType, Fault> EnableMotor(bool enable)
        {
            drive.EnableDriveRequest request = new drive.EnableDriveRequest();
            request.Enable = enable;

            return _drivePort.EnableDrive(request);
        }
        #endregion

        #region Robot state transition method

        //"Robot" -- get attention, start listening
        //"Turn Left" - rotate left 45 degrees
        //"Turn Right" - rotate right 45 degrees
        //"Turn Around" - rotate 180 degrees
        //"Go Forward" - begin moving forward (with avoiding obstacles)
        //"Backup" - translate backward
        //"Stop" - stop current behavior

        //*Behaviors using features of vision service : color object, face and hand gestures(left,right,both) detections
        //"Go There" - rotate after recognizing user's pointing hand gesture
        //"Follow Me" - start looking for and tracing the specified colored shirt

        //"Avoid" - avoid obstacle using LRF and retrieve the previous state

        /// <summary>
        /// Updates to the the BehaviorState
        /// </summary>
        public void SetRobotBehavior(RobotBehavior state)
        {
            _state.PreBehaviorState = state;
            _state.BehaviorState = state;
            stateCount = 0;
        }

        /// <summary>
        /// Updates to the LogicalState
        /// </summary>
        private void UpdateLogicalState()
        {
            //Continuous : Stop, Move
            //Discontinuous(_state.Countdown) : Turn, Translate, AdjustedMove
            if (_state.Countdown > 0)
            {
                _state.Countdown--;
            }
            else if (_state.LogicalState == LogicalStateType.Turn || _state.LogicalState == LogicalStateType.Translate || _state.LogicalState == LogicalStateType.AdjustedMove)
            {
                StopMoving();
                _state.LogicalState = LogicalStateType.Stop;
                _state.Countdown = 0;

                if (backupStarted)
                {
                    backupStarted = false;
                    _state.BumperPressed = BumperPressedType.None;
                }
            }
        }

        /// <summary>
        /// RandomTurn
        /// </summary>
        /// <param name="angle"></param>
        private void RandomTurn(int angle)
        {
            int heading = new Random().Next(-angle, angle);
            Turn(heading, TurnType.TwoWheelTurn);

            //LogInfo("Random Turning to: " + heading);
        }
        /// <summary>
        /// If the robot is mapping and there is sufficient open space directly ahead, enter this space.
        /// </summary>
        /// <param name="distance"></param>
        private void EnterOpenSpace(int distance)
        {
            int heading = FindBestFrom(_state.MostRecentLaserData, 0, _state.Velocity / 10, CorridorWidthMoving);

            if (distance > FreeDistance)
            {
                Move(MaximumForwardVelocity / 5);

                if (Math.Abs(heading) < 10)
                {
                    Turn(heading / 2, TurnType.OneWheelTurn);
                }
            }
            else if (distance > AwareOfObstacleDistance)
            {
                Move(MaximumForwardVelocity / 10);

                if (Math.Abs(heading) < 45)
                {
                    Turn(heading / 2, TurnType.OneWheelTurn);
                }
            }
            else
            {
                Turn(heading / 2, TurnType.TwoWheelTurn);
            }

            //LogInfo("EnterOpenSpace best heading to: " + heading);
        }

        /// <summary>
        /// If the robot is moving and an obstacle is too close, stop and map the environment for a way around it.
        /// </summary>
        /// <param name="distance"></param>
        private void AvoidCollision(int distance)
        {
            if (distance < ObstacleDistance)
            {
                if (_state.BehaviorState != RobotBehavior.Avoid)
                {
                    SayText(Resources.speakAvoidObstacle, 10);
                    _state.PreBehaviorState = _state.BehaviorState;
                }
                _state.BehaviorState = RobotBehavior.Avoid;
                int heading = FindBestFrom(_state.MostRecentLaserData, 0, _state.Velocity / 10, CorridorWidthMoving);

                Turn(heading, TurnType.TwoWheelTurn);
                //LogInfo(string.Format("AvoidCollision = {0} best heading {1}", distance, heading));
            }
            else if (_state.BehaviorState == RobotBehavior.Avoid)
            {
                if (_state.PreBehaviorState != RobotBehavior.Avoid)
                    _state.BehaviorState = _state.PreBehaviorState;
            }
        }

        bool backupStarted = false;
        bool BackupByBumper()
        {
            if (_state.BumperPressed != BumperPressedType.None)
            {
                if (backupStarted == false)
                {
                    backupStarted = true;
                    if (_state.BumperPressed == BumperPressedType.Front)
                        Translate(BackupDistance);
                    else
                        Translate(-BackupDistance);
                    //LogInfo(string.Format("BackupByTranslate = {0}", BackupDistance));
                }
            }
            return backupStarted;
        }

        int stateCount = 0;
        bool behaviorStarted = false;
        private void UpdateRobotState()
        {
            bool foundObj = _state.currentObjectResult.ObjectFound;
            int gazingDegree = (int)(((double)_state.currentObjectResult.XMean - 0.5f * (double)ImageWidth) / (double)ImageWidth * -90);
            int distance = FindNearestObstacleInCorridor(_state.MostRecentLaserData, CorridorWidthMapping, 45);

            // handle null exception of sicklrf.State
            //if (distance == 0)
            //    return;

            // handle bumper backup immediately
            if (BackupByBumper() == true)
                return;

            // AvoidCollision method updates its BehaviorState to avoid and backup the current state
            //if (_state.BehaviorState != RobotBehavior.Stop)
            //    AvoidCollision(distance);

            switch (_state.BehaviorState)
            {
                case RobotBehavior.Avoid:
                    break;

                case RobotBehavior.Stop:
                    //UpdateStop(foundObj, gazingDegree);
                    break;

                case RobotBehavior.TurnLeft:
                case RobotBehavior.TurnRight:
                case RobotBehavior.TurnAround:
                case RobotBehavior.Backup:

                    if (behaviorStarted == false)
                    {
                        behaviorStarted = true;

                        if (_state.BehaviorState == RobotBehavior.Backup)
                        {
                            Translate(BackupDistance * 2);
                        }
                        else
                        {
                            int turnAngle = 60;
                            if (_state.BehaviorState == RobotBehavior.TurnRight)
                                turnAngle = -60;
                            else if (_state.BehaviorState == RobotBehavior.TurnAround)
                                turnAngle = -180;

                            Turn(turnAngle, TurnType.TwoWheelTurn);
                        }

                    }
                    else if (behaviorStarted == true && _state.Countdown == 0)
                    {
                        behaviorStarted = false;
                        SetRobotBehavior(RobotBehavior.Stop);
                    }
                    break;

                case RobotBehavior.GoForward:
                    Move(70);
                    break;

                case RobotBehavior.FollowMe:
                    UpdateFollowMe(foundObj, gazingDegree, distance);
                    break;

                case RobotBehavior.GoThere:
                    UpdateGoThere();
                    break;
                default:
                    break;

            }

            if (stateCount % 10 == 0)
            {
                Console.WriteLine("state: " + _state.BehaviorState + " stateCount: " + stateCount);
                //LogInfo(string.Format("gazingDegree {0} looseCount {1}, _state.Countdown {2} distance{3}", gazingDegree, looseCount, _state.Countdown, distance));
            }

            stateCount++;
        }

        private void UpdateGoThere()
        {
            if (stateCount < 2)
            {
                _state.FaceResultValidTime = 0;
                _state.HandGestureResultValidTime = 0;

                _state.currentHandGestureResult.LeftHandGestureFound = false;
                _state.currentHandGestureResult.RightHandGestureFound = false;

                StopMoving();
            }

            // waits several seconds until gesture finish
            if (_state.HandGestureResultValidTime == 1)
            {
                if (_state.currentHandGestureResult.LeftHandGestureFound && _state.currentHandGestureResult.RightHandGestureFound)
                {
                    SayText(Resources.speakBothHandsBackup);
                    SetRobotBehavior(RobotBehavior.Backup);
                }
                else if (_state.currentHandGestureResult.RightHandGestureFound)
                {
                    SayText(Resources.speakRightHandTurningLeft);
                    SetRobotBehavior(RobotBehavior.TurnLeft);
                }
                else if (_state.currentHandGestureResult.LeftHandGestureFound)
                {
                    SayText(Resources.speakLeftHandTurningRight);
                    SetRobotBehavior(RobotBehavior.TurnRight);
                }
            }
        }

        private void UpdateStop(bool foundObj, int gazingDegree)
        {

            if (stateCount < 2)
            {
                _state.HandGestureResultValidTime = 0;
                _state.currentHandGestureResult.LeftHandGestureFound = false;
                _state.currentHandGestureResult.RightHandGestureFound = false;

                StopMoving();
            }

            if (foundObj && Math.Abs(gazingDegree) < 15)
            {
                StopMoving();
            }
            else if (foundObj)
            {
                Turn(gazingDegree / 5, TurnType.TwoWheelTurn);
            }

            if (_state.HandGestureResultValidTime == 1)
            {
                if (_state.currentHandGestureResult.LeftHandGestureFound && _state.currentHandGestureResult.RightHandGestureFound)
                {
                    SayText(Resources.speakBothHands);
                    _state.HandGestureResultValidTime = 0;
                    _state.currentHandGestureResult.LeftHandGestureFound = false;
                    _state.currentHandGestureResult.RightHandGestureFound = false;
                }
                else if (_state.currentHandGestureResult.RightHandGestureFound)
                {
                    SayText(Resources.speakRightHand);
                    _state.HandGestureResultValidTime = 0;
                    _state.currentHandGestureResult.LeftHandGestureFound = false;
                    _state.currentHandGestureResult.RightHandGestureFound = false;
                }
                else if (_state.currentHandGestureResult.LeftHandGestureFound)
                {
                    SayText(Resources.speakLeftHand);
                    _state.HandGestureResultValidTime = 0;
                    _state.currentHandGestureResult.LeftHandGestureFound = false;
                    _state.currentHandGestureResult.RightHandGestureFound = false;

                }
            }

            if (_state.FaceResultValidTime > 8 && foundObj)
            {
                SayText(Resources.speakFoundFace, 50);
            }
        }

        int objectLossCount = 0;
        private void UpdateFollowMe(bool foundObj, int gazingDegree, int distance)
        {
            int goalDistanceSize = 2500;

            // range of gazingDegree = [ -45 ~ 0 ~ 45 ] 
            if (foundObj && Math.Abs(gazingDegree) < 10 &&
                (_state.currentObjectResult.Area > goalDistanceSize || _state.FaceResultValidTime > 5))
            {
                //LogInfo(string.Format("Got to Goal = {0}", currentObjectResult.Area));
                StopMoving();
            }
            else if (foundObj)
            {
                SayText(Resources.speakFoundObject, 30);

                if (gazingDegree > -8 && gazingDegree < 8)
                    Translate(70);
                else if (gazingDegree > -20 && gazingDegree < 20 && _state.currentObjectResult.Area < goalDistanceSize)
                    AdjustedMove(gazingDegree, 60);
                else
                    Turn(gazingDegree , TurnType.OneWheelTurn);

                objectLossCount = 0;
            }
            else if (!foundObj && _state.Countdown == 0)
            {
                SayText(Resources.speakSearching, 20);

                if (objectLossCount < 30)
                    Turn(gazingDegree , TurnType.TwoWheelTurn);
                else if (objectLossCount >= 30)
                {
                    if (objectLossCount % 5 == 0 && objectLossCount < 100)
                    {
                        if (new Random().Next(10) < 1)
                            RandomTurn(180);
                        //else
                        //    EnterOpenSpace(distance);
                    }
                    if (objectLossCount % 5 == 0)
                    {
                        if (new Random().Next(10) < 3)
                            RandomTurn(360);
                        //else
                        //    EnterOpenSpace(distance);
                    }

                }

                objectLossCount++;
            }
        }

        private void SayText(string text, int freq)
        {
            if (stateCount % freq == 0)
            {
                SpawnIterator(text, SayTextIterator);
            }
        }

        private void SayText(string text)
        {
            SpawnIterator(text, SayTextIterator);
        }

        IEnumerator<ITask> SayTextIterator(string text)
        {
            // disable speech recognition for an echo canceling
            DisableSpeechUpdate request = new DisableSpeechUpdate(true);
            _disableSpeechPort.Post(request);
            LogInfo("Speech Recognition Disabled");

            yield return Arbiter.Choice(
                request.ResponsePort,
                delegate(DefaultUpdateResponseType response) { },
                delegate(Fault fault) { }
            );


            _winMsgPort.SendMessageString("TTSAgent",text);
            /*
            winMsg.WinMsgCopyData winMsgRequest = new winMsg.WinMsgCopyData();
            winMsgRequest.ProcessName = "TTSAgent";
            winMsgRequest.StringData = text;
            _winMsgPort.Post(new winMsg.SendMessageString(winMsgRequest));
            */
            // recover speech recognition flag
            Activate(Arbiter.Receive(false, TimeoutPort(text.Length * 100),
                   delegate(DateTime ts)
                   {
                       LogInfo("Speech Recognition Enabled");
                       _disableSpeechPort.Post(new DisableSpeechUpdate(false));
                   }
            ));
            yield break;
        }

        void OnSpeechRecognition(winMsg.ReceiveMessage recognition)
        {
            if (disableSpeech == true)
            {
                recognition.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                return;
            }
            LogInfo(LogGroups.Console,"Got Speech Command = " + recognition.Body.StringData);

            // keyword spotting by a word 'Robot'
            if (recognition.Body.WParam == 1)
            {


                if (recognition.Body.StringData  == "κ")
                {
                    // set SpeechCommandValidTime to 20 (150 mSec x 20 = 3 Sec)
                    _state.SpeechCommandValidTime = 20;

                    //LogInfo("Recognized <Robot> = ready for real command");
                    recognition.ResponsePort.Post(DefaultUpdateResponseType.Instance);
                    return;
                }
                else if (_state.SpeechCommandValidTime > 0)
                {
                    //LogInfo("Got Speech Command = " + recognition.Body.StringData);
                    switch (recognition.Body.StringData)
                    {
                        case "":
                        case "׸":
                            SayText(recognition.Body.StringData);
                            SetRobotBehavior(RobotBehavior.Stop);
                            break;

                        case "λ":
                            SayText(Resources.speakHello);
                            break;
                        case "ε":
                            SayText(Resources.speakTurnLeft);
                            SetRobotBehavior(RobotBehavior.TurnLeft);
                            break;
                        case "ε":
                            SayText(Resources.speakTurnRight);
                            SetRobotBehavior(RobotBehavior.TurnRight);
                            break;
                        case "ڷε":
                            SayText(Resources.speakTurnAround);
                            SetRobotBehavior(RobotBehavior.TurnAround);
                            break;
                        case "ΰ":
                            SayText(Resources.speakForward);
                            SetRobotBehavior(RobotBehavior.GoForward);
                            break;
                        case "":
                            SetRobotBehavior(RobotBehavior.FollowMe);
                            break;
                        case "ڷι":
                            SayText(Resources.speakBackup);
                            SetRobotBehavior(RobotBehavior.Backup);
                            break;
                        case "":
                            SayText(Resources.speakShowMe);
                            SetRobotBehavior(RobotBehavior.GoThere);
                            break;
                    }
                }
                else
                {
                    SayText(Resources.speakCallMe, 5);
                }
            }


            recognition.ResponsePort.Post(DefaultUpdateResponseType.Instance);
        }
        #endregion

    }
}
