//-----------------------------------------------------------------------
//  This file is part of the Microsoft Robotics Studio Code Samples.
// 
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//  $File: FollowerState.cs $ $Revision: 5 $
//-----------------------------------------------------------------------

using Microsoft.Dss.Core.Attributes;
using System;
using System.Collections.Generic;
using sicklrf = Microsoft.Robotics.Services.Sensors.SickLRF.Proxy;
using drive = Microsoft.Robotics.Services.Drive.Proxy;
using vision = Microsoft.Robotics.Services.Sample.SimpleVision.Proxy;
using System.ComponentModel;

namespace Microsoft.Robotics.Services.Sample.Follower
{
    /// <summary>
    /// FollowerState
    /// </summary>
    [DataContract()]
    public class FollowerState
    {
        #region private fields
        
        private drive.DriveDifferentialTwoWheelState _driveState;
        private int _velocity;

        private DateTime _mostRecentLaser = DateTime.Now;
        private sicklrf.State _mostRecentLaserData;

        private int _countdown;
        private LogicalStateType _logicalState = LogicalStateType.Stop;

        private RobotBehavior _preBehaviorState = RobotBehavior.Stop;
        private RobotBehavior _behaviorState = RobotBehavior.Stop;

        private BumperPressedType _bumperPressed = BumperPressedType.None;

        #endregion

        #region data members
        /// <summary>
        /// DriveState
        /// </summary>
        [DataMember(IsRequired=false)]
        public drive.DriveDifferentialTwoWheelState DriveState
        {
            get { return _driveState; }
            set { _driveState = value; }
        }
        /// <summary>
        /// Driving Velocity
        /// </summary>
        [DataMember]
        public int Velocity
        {
            get { return _velocity; }
            set { _velocity = value; }
        }
        /// <summary>
        /// MostRecentLaserData
        /// </summary>
        [DataMember]
        public sicklrf.State MostRecentLaserData
        {
            get { return _mostRecentLaserData; }
            set { _mostRecentLaserData = value; }
        }
        /// <summary>
        /// MostRecentLaserTime
        /// </summary>
        [DataMember]
        public DateTime MostRecentLaserTime
        {
            get { return _mostRecentLaser; }
            set { _mostRecentLaser = value; }
        }
        /// <summary>
        /// Countdown
        /// </summary>
        [DataMember]
        public int Countdown
        {
            get { return _countdown; }
            set { _countdown = value; }
        }
        /// <summary>
        /// LogicalState of Robot
        /// </summary>
        [DataMember]
        public LogicalStateType LogicalState
        {
            get { return _logicalState; }
            set { _logicalState = value; }
        }        
        /// <summary>
        /// Previous BehaviorState of Robot
        /// </summary>
        [DataMember]
        public RobotBehavior PreBehaviorState
        {
            get { return _preBehaviorState; }
            set { _preBehaviorState = value; }
        }
        /// <summary>
        /// Current BehaviorState of Robot
        /// </summary>
        [DataMember]
        public RobotBehavior BehaviorState
        {
            get { return _behaviorState; }
            set { _behaviorState = value; }
        }
        /// <summary>
        /// BumperPressed
        /// </summary>
        public BumperPressedType BumperPressed
        {
            get { return _bumperPressed; }
            set { _bumperPressed = value; }
        }
        /// <summary>
        /// current Object Detection Result
        /// </summary>
        [DataMember]
        public vision.ObjectResult currentObjectResult;
        /// <summary>
        /// current Face Detection Result
        /// </summary>
        [DataMember]
        public vision.FaceResult currentFaceResult;
        /// <summary>
        /// current HandGesture Detection Result
        /// </summary>
        [DataMember]
        public vision.HandGestureResult currentHandGestureResult;
        /// <summary>
        /// TimerIntervalInMs for TimerHandler
        /// </summary>
        [DataMember]
        public int TimerIntervalInMs;
        /// <summary>
        /// Valid Time of Face Detection Result
        /// </summary>
        [DataMember]
        public int FaceResultValidTime;
        /// <summary>
        /// Valid Time of HandGesture Detectio Result
        /// </summary>
        [DataMember]
        public int HandGestureResultValidTime;
        /// <summary>
        /// Valid Time of Speech Command Recognition
        /// </summary>
        [DataMember]
        public int SpeechCommandValidTime;

        #endregion
    }

    /// <summary>
    /// Request Robot Behavior
    /// </summary>
    [DataContract, DataMemberConstructor]
    public class RequestRobotBehavior
    {
        /// <summary>
        /// The Robot Behavior Requested
        /// </summary>
        [DataMember]
        [Description("Indicates the behavior requested.")]
        public RobotBehavior Behavior;

        /// <summary>
        /// Request Robot Behavior
        /// </summary>
        public RequestRobotBehavior() { }

        /// <summary>
        /// Request Robot Behavior
        /// </summary>
        /// <param name="behavior"></param>
        public RequestRobotBehavior(RobotBehavior behavior) { this.Behavior = behavior; }
    }

    /// <summary>
    /// LogicalStateType for robot movement types definition
    /// </summary>
    [DataContract]
    public enum LogicalStateType
    {
        /// <summary>
        /// Stop : Continuous
        /// </summary>
        Stop,
        /// <summary>
        /// Move forward : Continuous
        /// </summary>
        Move,
        /// <summary>
        /// Turn : Discontinuous(_state.Countdown)
        /// </summary>
        Turn,
        /// <summary>
        /// Translate : Discontinuous(_state.Countdown)
        /// </summary>
        Translate,
        /// <summary>
        /// AdjustedMove : Discontinuous(_state.Countdown)
        /// </summary>
        AdjustedMove,
    }

    /// <summary>
    /// RobotBehavior for robot behavior types definition
    /// </summary>
    [DataContract]
    [Description("Identifies a specific behavior.")]
    public enum RobotBehavior
    {
        /// <summary>
        /// Stop current behavior and Ready (the default Action)
        /// </summary>
        Stop = 0,
        /// <summary>
        /// Avoid obstacle using LRF and retrieve the previous state
        /// </summary>
        Avoid,
        /// <summary>
        /// Rotate left 45 degrees
        /// </summary>
        TurnLeft,
        /// <summary>
        /// Rotate right 45 degrees
        /// </summary>
        TurnRight,
        /// <summary>
        /// Rotate 180 degrees
        /// </summary>
        TurnAround,
        /// <summary>
        /// Begin moving forward with avoiding obstacles
        /// </summary>
        GoForward,
        /// <summary>
        /// Translate backward
        /// </summary>
        Backup,
        /// <summary>
        /// Start looking for and tracing the specified colored shirt
        /// </summary>
        FollowMe,
        /// <summary>
        /// Rotate after recognizing user's pointing hand gesture
        /// </summary>
        GoThere,
    }

    /// <summary>
    /// TurnType for turn types definition
    /// </summary>
    [DataContract]
    public enum TurnType
    {
        /// <summary>
        /// use single wheel
        /// </summary>
        OneWheelTurn,
        /// <summary>
        /// use both wheel
        /// </summary>
        TwoWheelTurn,
    }

    /// <summary>
    /// BumperPressedType for bumper press types definition
    /// </summary>
    [DataContract]
    public enum BumperPressedType
    {
        /// <summary>
        /// None of bumpers pressed
        /// </summary>
        None,
        /// <summary>
        /// Front bumper pressed
        /// </summary>
        Front,
        /// <summary>
        /// Rear bumper pressed
        /// </summary>
        Rear,
    }

}